@@ -21,31 +21,33 @@
 
 # Imports
 import ecdsa
-from ecdsa.curves                 import SECP256k1
-from ecdsa.ecdsa                  import generator_secp256k1, int_to_string, string_to_int
-from bip_utils.bip.bip32_ex       import Bip32KeyError, Bip32PathError
-from bip_utils.bip.bip32_utils    import Bip32Utils
-from bip_utils.bip.bip32_path     import Bip32PathParser
-from bip_utils.bip.bip32_key_ser  import Bip32KeyDeserializer
-from bip_utils.bip.bip_keys       import BipPrivateKey, BipPublicKey
-from bip_utils.conf.bip_coin_conf import Bip32Conf
-from bip_utils.utils              import CryptoUtils
+from typing import Tuple
+from ecdsa.curves import SECP256k1
+from ecdsa.ecdsa import generator_secp256k1
+from bip_utils.bip.bip32_ex import Bip32KeyError, Bip32PathError
+from bip_utils.bip.bip32_utils import Bip32Utils
+from bip_utils.bip.bip32_path import Bip32PathParser
+from bip_utils.bip.bip32_key_ser import Bip32KeyDeserializer
+from bip_utils.bip.bip_keys import BipPrivateKey, BipPublicKey
+from bip_utils.conf import Bip32Conf, KeyNetVersions
+from bip_utils.ecc import EcdsaPublicKey, EcdsaPrivateKey
+from bip_utils.utils import CryptoUtils, ConvUtils
 
 
 class Bip32Const:
     """ Class container for BIP32 constants. """
 
     # SECP256k1 curve order
-    CURVE_ORDER          = generator_secp256k1.order()
+    CURVE_ORDER: int = generator_secp256k1.order()
 
     # Fingerprint length in bytes
-    FINGERPRINT_BYTE_LEN = 4
+    FINGERPRINT_BYTE_LEN: int = 4
     # Fingerprint of master key
-    MASTER_FINGERPRINT   = b"\x00\x00\x00\x00"
+    MASTER_FINGERPRINT: bytes = b"\x00\x00\x00\x00"
     # Minimum length in bits for seed
-    SEED_MIN_BIT_LEN     = 128
+    SEED_MIN_BIT_LEN: int = 128
     # HMAC key for generating master key
-    MASTER_KEY_HMAC_KEY  = b"Bitcoin seed"
+    MASTER_KEY_HMAC_KEY: bytes = b"Bitcoin seed"
 
 
 class Bip32:
@@ -58,7 +60,8 @@ class Bip32:
     #
 
     @staticmethod
-    def FromSeed(seed_bytes, key_net_ver = Bip32Conf.KEY_NET_VER.Main()):
+    def FromSeed(seed_bytes: bytes,
+                 key_net_ver: KeyNetVersions = Bip32Conf.KEY_NET_VER.Main()) -> 'Bip32':
         """ Create a Bip32 object from the specified seed (e.g. BIP39 seed).
 
         Args:
@@ -80,10 +83,14 @@ class Bip32:
         # Compute HMAC
         hmac = CryptoUtils.HmacSha512(Bip32Const.MASTER_KEY_HMAC_KEY, seed_bytes)
         # Create BIP32 by splitting the HMAC into two 32-byte sequences
-        return Bip32(secret = hmac[:32], chain = hmac[32:], key_net_ver = key_net_ver)
+        return Bip32(secret=hmac[:32],
+                     chain=hmac[32:],
+                     key_net_ver=key_net_ver)
 
     @staticmethod
-    def FromSeedAndPath(seed_bytes, path, key_net_ver = Bip32Conf.KEY_NET_VER.Main()):
+    def FromSeedAndPath(seed_bytes: bytes,
+                        path: str,
+                        key_net_ver: KeyNetVersions = Bip32Conf.KEY_NET_VER.Main()) -> 'Bip32':
         """ Create a Bip32 object from the specified seed (e.g. BIP39 seed) and path.
 
         Args:
@@ -115,7 +122,8 @@ class Bip32:
         return bip32_ctx
 
     @staticmethod
-    def FromExtendedKey(key_str, key_net_ver = Bip32Conf.KEY_NET_VER.Main()):
+    def FromExtendedKey(key_str: str,
+                        key_net_ver: KeyNetVersions = Bip32Conf.KEY_NET_VER.Main()) -> 'Bip32':
         """ Create a Bip32 object from the specified extended key.
 
         Args:
@@ -151,40 +159,41 @@ class Bip32:
         # If public key, recover the complete key from the compressed one
         else:
             try:
-                secret = ecdsa.VerifyingKey.from_string(secret, curve = SECP256k1)
-            except ecdsa.keys.MalformedPointError:
-                raise Bip32KeyError("Invalid extended public key (malformed point)")
+                secret = ecdsa.VerifyingKey.from_string(secret, curve=SECP256k1)
+            except ecdsa.keys.MalformedPointError as ex:
+                raise Bip32KeyError("Invalid extended public key (malformed point)") from ex
 
-        return Bip32(
-            secret      = secret,
-            chain       = chain,
-            depth       = depth,
-            index       = index,
-            fprint      = fprint,
-            is_public   = is_public,
-            key_net_ver = key_net_ver)
+        return Bip32(secret=secret,
+                     chain=chain,
+                     depth=depth,
+                     index=index,
+                     fprint=fprint,
+                     is_public=is_public,
+                     key_net_ver=key_net_ver)
 
     #
     # Public methods
     #
 
     def __init__(self,
-                 secret,
-                 chain,
-                 depth       = 0,
-                 index       = 0,
-                 fprint      = Bip32Const.MASTER_FINGERPRINT,
-                 is_public   = False,
-                 key_net_ver = Bip32Conf.KEY_NET_VER.Main()):
+                 secret: bytes,
+                 chain: bytes,
+                 depth: int = 0,
+                 index: int = 0,
+                 fprint: bytes = Bip32Const.MASTER_FINGERPRINT,
+                 is_public: bool = False,
+                 key_net_ver: KeyNetVersions = Bip32Conf.KEY_NET_VER.Main()) -> None:
         """ Construct class from secret and chain.
 
         Args:
             secret (bytes)                               : Source bytes to generate the keypair
             chain (bytes)                                : 32-byte representation of the chain code
-            depth (int, optional)                        : Child depth, parent increments its own by one when assigning this (default: 0)
+            depth (int, optional)                        : Child depth, parent increments its own by one when
+                                                           assigning this (default: 0)
             index (int, optional)                        : Child index (default: 0)
             fprint (bytes, optional)                     : Parent fingerprint (default: 0)
-            is_public (bool, optional)                   : If true, this keypair will only contain a public key and can only create a public key chain  (default: false)
+            is_public (bool, optional)                   : If true, this keypair will only contain a public key and can
+                                                           only create a public key chain  (default: false)
             key_net_ver (KeyNetVersions object, optional): Key net version object (Bip32 main net version by default)
 
         Raises:
@@ -194,23 +203,24 @@ class Bip32:
         if not is_public:
             # Check private key validity
             try:
-                self.m_key = ecdsa.SigningKey.from_string(secret, curve = SECP256k1)
-            except ecdsa.keys.MalformedPointError:
-                raise Bip32KeyError("Invalid private key (malformed point)")
+                self.m_key = ecdsa.SigningKey.from_string(secret, curve=SECP256k1)
+            except ecdsa.keys.MalformedPointError as ex:
+                raise Bip32KeyError("Invalid private key (malformed point)") from ex
             # Get verifying key
             self.m_ver_key = self.m_key.get_verifying_key()
         else:
-            self.m_key     = None
+            self.m_key = None
             self.m_ver_key = secret
 
-        self.m_is_public     = is_public
-        self.m_chain         = chain
-        self.m_depth         = depth
-        self.m_index         = index
+        self.m_is_public = is_public
+        self.m_chain = chain
+        self.m_depth = depth
+        self.m_index = index
         self.m_parent_fprint = fprint
-        self.m_key_net_ver   = key_net_ver
+        self.m_key_net_ver = key_net_ver
 
-    def ChildKey(self, index):
+    def ChildKey(self,
+                 index: int) -> 'Bip32':
         """ Create and return a child key of the current one at the specified index.
         The index shall be hardened using HardenIndex method to use the private derivation algorithm.
 
@@ -225,7 +235,8 @@ class Bip32:
         """
         return self.__CkdPriv(index) if not self.m_is_public else self.__CkdPub(index)
 
-    def DerivePath(self, path):
+    def DerivePath(self,
+                   path: str) -> 'Bip32':
         """ Derive children keys from the specified path.
 
         Args:
@@ -252,12 +263,12 @@ class Bip32:
 
         return bip32_obj
 
-    def ConvertToPublic(self):
+    def ConvertToPublic(self) -> None:
         """ Convert a private Bip32 object into a public one. """
-        self.m_key       = None
+        self.m_key = None
         self.m_is_public = True
 
-    def IsPublicOnly(self):
+    def IsPublicOnly(self) -> bool:
         """ Get if it's public-only.
 
         Returns:
@@ -265,29 +276,29 @@ class Bip32:
         """
         return self.m_is_public
 
-    def EcdsaPrivateKey(self):
-        """ Return the ECDSA private key object.
+    def EcdsaPrivateKey(self) -> EcdsaPrivateKey:
+        """ Return ECC private key object.
 
-        Return:
-            ecdsa.SigningKey object: ecdsa.SigningKey object
+        Returns:
+            EcdsaPrivateKey object: EcdsaPrivateKey object
 
         Raises:
             Bip32KeyError: If internal key is public-only
         """
         if self.m_is_public:
             raise Bip32KeyError("Public-only deterministic keys have no private half")
-        return self.m_key
+        return EcdsaPrivateKey(self.m_key)
 
-    def EcdsaPublicKey(self):
-        """ Return the ECDSA public key object.
+    def EcdsaPublicKey(self) -> EcdsaPublicKey:
+        """ Return ECC public key object.
 
-        Return:
-            ecdsa.VerifyingKey object: ecdsa.VerifyingKey object
+        Returns:
+            EcdsaPublicKey object: EcdsaPublicKey object
         """
-        return self.m_ver_key
+        return EcdsaPublicKey(self.m_ver_key)
 
-    def PrivateKey(self):
-        """ Return private key bytes.
+    def PrivateKey(self) -> BipPrivateKey:
+        """ Return private key object.
 
         Returns:
             BipPrivateKey object: BipPrivateKey object
@@ -295,27 +306,35 @@ class Bip32:
         Raises:
             Bip32KeyError: If internal key is public-only
         """
-        if self.m_is_public:
-            raise Bip32KeyError("Public-only deterministic keys have no private half")
-        return BipPrivateKey(self)
+        return BipPrivateKey(self.EcdsaPrivateKey(),
+                             self.KeyNetVersions(),
+                             self.Depth(),
+                             self.ParentFingerPrint(),
+                             self.Index(),
+                             self.Chain())
 
-    def PublicKey(self):
-        """ Return public key bytes.
+    def PublicKey(self) -> BipPublicKey:
+        """ Return public key object.
 
         Returns:
             BipPublicKey object: BipPublicKey object
         """
-        return BipPublicKey(self)
+        return BipPublicKey(self.EcdsaPublicKey(),
+                            self.KeyNetVersions(),
+                            self.Depth(),
+                            self.ParentFingerPrint(),
+                            self.Index(),
+                            self.Chain())
 
-    def KeyNetVersions(self):
+    def KeyNetVersions(self) -> KeyNetVersions:
         """ Get key net versions.
 
         Returns:
-            KeyKeyNetVersions object: KeyKeyNetVersions object
+            KeyNetVersions object: KeyNetVersions object
         """
         return self.m_key_net_ver
 
-    def Depth(self):
+    def Depth(self) -> int:
         """ Get current depth.
 
         Returns:
@@ -323,7 +342,7 @@ class Bip32:
         """
         return self.m_depth
 
-    def Index(self):
+    def Index(self) -> int:
         """ Get current index.
 
         Returns:
@@ -331,7 +350,7 @@ class Bip32:
         """
         return self.m_index
 
-    def Chain(self):
+    def Chain(self) -> bytes:
         """ Get current chain.
 
         Returns:
@@ -339,15 +358,15 @@ class Bip32:
         """
         return self.m_chain
 
-    def KeyIdentifier(self):
+    def KeyIdentifier(self) -> bytes:
         """ Get key identifier.
 
         Returns:
             bytes: Key identifier bytes
         """
-        return CryptoUtils.Hash160(self.PublicKey().RawCompressed().ToBytes())
+        return CryptoUtils.Hash160(self.EcdsaPublicKey().RawCompressed().ToBytes())
 
-    def FingerPrint(self):
+    def FingerPrint(self) -> bytes:
         """ Get key fingerprint.
 
         Returns:
@@ -355,7 +374,7 @@ class Bip32:
         """
         return self.KeyIdentifier()[:Bip32Const.FINGERPRINT_BYTE_LEN]
 
-    def ParentFingerPrint(self):
+    def ParentFingerPrint(self) -> bytes:
         """ Get parent fingerprint.
 
         Returns:
@@ -367,7 +386,8 @@ class Bip32:
     # Private methods
     #
 
-    def __CkdPriv(self, index):
+    def __CkdPriv(self,
+                  index: int) -> 'Bip32':
         """ Create a child key of the specified index.
 
         Args:
@@ -385,32 +405,33 @@ class Bip32:
 
         # Data for HMAC
         if Bip32Utils.IsHardenedIndex(index):
-            data = b"\x00" + self.m_key.to_string() + index_bytes
+            data = b"\x00" + self.EcdsaPrivateKey().Raw().ToBytes() + index_bytes
         else:
-            data = self.PublicKey().RawCompressed().ToBytes() + index_bytes
+            data = self.EcdsaPublicKey().RawCompressed().ToBytes() + index_bytes
 
         # Compute HMAC halves
         i_l, i_r = self.__HmacHalves(data)
 
         # Construct new key secret from i_l and current private key
-        i_l_int = string_to_int(i_l)
-        key_int = string_to_int(self.m_key.to_string())
+        i_l_int = ConvUtils.BytesToInteger(i_l)
+        key_int = ConvUtils.BytesToInteger(self.m_key.to_string())
         new_key_int = (i_l_int + key_int) % Bip32Const.CURVE_ORDER
 
         # Convert to string and left pad with zeros
-        secret = int_to_string(new_key_int)
+        secret = ConvUtils.IntegerToBytes(new_key_int)
         secret = b"\x00" * (32 - len(secret)) + secret
 
         # Construct and return a new Bip32 object
-        return Bip32(secret      = secret,
-                     chain       = i_r,
-                     depth       = self.m_depth + 1,
-                     index       = index,
-                     fprint      = self.FingerPrint(),
-                     is_public   = False,
-                     key_net_ver = self.m_key_net_ver)
+        return Bip32(secret=secret,
+                     chain=i_r,
+                     depth=self.m_depth + 1,
+                     index=index,
+                     fprint=self.FingerPrint(),
+                     is_public=False,
+                     key_net_ver=self.m_key_net_ver)
 
-    def __CkdPub(self, index):
+    def __CkdPub(self,
+                 index: int) -> 'Bip32':
         """ Create a publicly derived child key of the specified index.
 
         Args:
@@ -428,31 +449,33 @@ class Bip32:
             raise Bip32KeyError("Public child derivation cannot be used to create a hardened child key")
 
         # Data for HMAC, same of __CkdPriv() for public child key
-        data = self.PublicKey().RawCompressed().ToBytes() + index.to_bytes(4, "big")
+        data = self.EcdsaPublicKey().RawCompressed().ToBytes() + index.to_bytes(4, "big")
 
         # Get HMAC of data
         i_l, i_r = self.__HmacHalves(data)
 
         # Construct curve point i_l*G+K
-        point = string_to_int(i_l) * generator_secp256k1 + self.m_ver_key.pubkey.point
+        point = ConvUtils.BytesToInteger(i_l) * generator_secp256k1 + self.m_ver_key.pubkey.point
 
         # Try to construct public key from the curve point
         try:
-            k_i = ecdsa.VerifyingKey.from_public_point(point, curve = SECP256k1)
-        except:
-            raise Bip32KeyError("Computed public child key is not valid, very unlucky index")
+            k_i = ecdsa.VerifyingKey.from_public_point(point, curve=SECP256k1)
+        except ecdsa.keys.MalformedPointError as ex:
+            raise Bip32KeyError("Computed public child key is not valid, very unlucky index") from ex
 
         # Construct and return a new Bip32 object
-        return Bip32(secret      = k_i,
-                     chain       = i_r,
-                     depth       = self.m_depth + 1,
-                     index       = index,
-                     fprint      = self.FingerPrint(),
-                     is_public   = True,
-                     key_net_ver = self.m_key_net_ver)
+        return Bip32(secret=k_i,
+                     chain=i_r,
+                     depth=self.m_depth + 1,
+                     index=index,
+                     fprint=self.FingerPrint(),
+                     is_public=True,
+                     key_net_ver=self.m_key_net_ver)
 
-    def __HmacHalves(self, data_bytes):
-        """ Calculate the HMAC-SHA512 of input data using the chain code as key and returns a tuple of the left and right halves of the HMAC.
+    def __HmacHalves(self,
+                     data_bytes: bytes) -> Tuple[bytes, bytes]:
+        """ Calculate the HMAC-SHA512 of input data using the chain code as key and returns a tuple
+        of the left and right halves of the HMAC.
 
         Args:
             data_bytes (bytes): Data bytes
@@ -463,4 +486,4 @@ class Bip32:
 
         # Use chain as HMAC key
         hmac = CryptoUtils.HmacSha512(self.m_chain, data_bytes)
-        return (hmac[:32], hmac[32:])
+        return hmac[:32], hmac[32:]
